/// @file   DpmMySql.cpp
/// @brief  MySQL DPM Implementation.
/// @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
#include <dmlite/cpp/dmlite.h>
#include <dmlite/cpp/utils/urls.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>

#include "DpmMySql.h"
#include "Queries.h"
#include "MySqlFactories.h"

using namespace dmlite;



MySqlPoolManager::MySqlPoolManager(DpmMySqlFactory* factory,
                                   const std::string& dpmDb) throw (DmException):
      stack_(0x00), dpmDb_(dpmDb), factory_(factory)
{
  this->conn_ = factory->getConnection();
}



MySqlPoolManager::~MySqlPoolManager()
{
  this->factory_->releaseConnection(this->conn_);
}



std::string MySqlPoolManager::getImplId() const throw ()
{
  return "mysql_pool_manager";
}



void MySqlPoolManager::setStackInstance(StackInstance* si) throw (DmException)
{
  this->stack_ = si;
}



void MySqlPoolManager::setSecurityContext(const SecurityContext* ctx) throw (DmException)
{
  this->secCtx_ = ctx;
}



std::vector<Pool> MySqlPoolManager::getPools(PoolAvailability availability) throw (DmException)
{
  Pool pool;
  std::vector<Pool> pools;
  char poolName[kPoolNameMax], poolType[kPoolTypeMax], poolMeta[kPoolMetaMax];
  
  // Get all
  Statement stmt(this->conn_, this->dpmDb_, STMT_GET_POOLS);

  stmt.execute();

  stmt.bindResult(0, poolName, sizeof(poolName));
  stmt.bindResult(1, poolType, sizeof(poolType));
  stmt.bindResult(2, poolMeta, sizeof(poolMeta));

  while (stmt.fetch()) {
    pool.clear();
    pool.name = poolName;
    pool.type = poolType;
    pool.deserialize(poolMeta);
    pools.push_back(pool);
  }

  // Filter
  if (availability == kAny)
    return pools;
  else {
    std::vector<Pool> filtered;
  
    for (unsigned i = 0; i < pools.size(); ++i) {
      PoolHandler* handler = this->stack_->getPoolDriver(pools[i].type)->
                               createPoolHandler(pools[i].name);
    
      bool isAvailable = handler->poolIsAvailable(availability == kForWrite ||
                                                  availability == kForBoth);
      
      if ((availability == kNone && !isAvailable) ||
          (availability != kNone && isAvailable))
        filtered.push_back(pools[i]);
    
      delete handler;
    }
    return filtered;
  }
}



Pool MySqlPoolManager::getPool(const std::string& poolname) throw (DmException)
{
  Pool pool;
  char poolName[kPoolNameMax], poolType[kPoolTypeMax], poolMeta[kPoolMetaMax];
  
  Statement stmt(this->conn_, this->dpmDb_, STMT_GET_POOL);

  stmt.bindParam(0, poolname);
  stmt.execute();

  stmt.bindResult(0, poolName, sizeof(poolName));
  stmt.bindResult(1, poolType, sizeof(poolType));
  stmt.bindResult(2, poolMeta, sizeof(poolMeta));

  if (!stmt.fetch())
    throw DmException(DM_NO_SUCH_POOL, "Pool %s not found", poolname.c_str());

  pool.name = poolName;
  pool.type = poolType;
  pool.deserialize(poolMeta);
  
  return pool;
}



void MySqlPoolManager::newPool(const Pool& pool) throw (DmException)
{
  PoolDriver* driver = this->stack_->getPoolDriver(pool.type);
  
  // Call the driver before
  driver->toBeCreated(pool);

  // Filesystem is the legacy pooltype, so the driver creates itself
  // the entry. Not nice, but we need to keep compatibility, and
  // since the running DPM daemon wouldn't see this until it is
  // restarted, let it do it itself.
  if (pool.type != "filesystem") {
    Statement stmt(this->conn_, this->dpmDb_, STMT_INSERT_POOL);

    std::vector<boost::any> groups = pool.getVector("groups");
    std::ostringstream gids;

    if (groups.size() > 0) {
      unsigned i;
      for (i = 0; i < groups.size() - 1; ++i)
        gids << Extensible::anyToUnsigned(groups[i]) << ",";
      gids << Extensible::anyToUnsigned(groups[i]);
    }
    else {
      gids << "0";
    }

    stmt.bindParam( 0, pool.name);
    stmt.bindParam( 1, pool.getLong("defsize"));
    stmt.bindParam( 2, pool.getLong("gc_start_thresh"));
    stmt.bindParam( 3, pool.getLong("gc_stop_thresh"));
    stmt.bindParam( 4, pool.getLong("def_lifetime"));
    stmt.bindParam( 5, pool.getLong("defpintime"));
    stmt.bindParam( 6, pool.getLong("max_lifetime"));
    stmt.bindParam( 7, pool.getLong("maxpintime"));
    stmt.bindParam( 8, pool.getString("fss_policy"));
    stmt.bindParam( 9, pool.getString("gc_policy"));
    stmt.bindParam(10, pool.getString("mig_policy"));
    stmt.bindParam(11, pool.getString("rs_policy"));
    stmt.bindParam(12, gids.str());
    stmt.bindParam(13, pool.getString("ret_policy"));
    stmt.bindParam(14, pool.getString("s_type"));
    stmt.bindParam(15, pool.type);
    stmt.bindParam(16, pool.serialize());

    stmt.execute();
  }
  
  // Let the driver know
  driver->justCreated(pool);
}



void MySqlPoolManager::updatePool(const Pool& pool) throw (DmException)
{
  PoolDriver* driver = this->stack_->getPoolDriver(pool.type);
  
  // Update the db
  Statement stmt(this->conn_, this->dpmDb_, STMT_UPDATE_POOL);
  
  std::vector<boost::any> groups = pool.getVector("groups");
  std::ostringstream gids;
  
  if (groups.size() > 0) {
    unsigned i;
    for (i = 0; i < groups.size() - 1; ++i)
      gids << Extensible::anyToUnsigned(groups[i]) << ",";
    gids << Extensible::anyToUnsigned(groups[i]);
  }
  else {
    gids << "0";
  }
  
  stmt.bindParam( 0, pool.getLong("defsize"));
  stmt.bindParam( 1, pool.getLong("gc_start_thresh"));
  stmt.bindParam( 2, pool.getLong("gc_stop_thresh"));
  stmt.bindParam( 3, pool.getLong("def_lifetime"));
  stmt.bindParam( 4, pool.getLong("defpintime"));
  stmt.bindParam( 5, pool.getLong("max_lifetime"));
  stmt.bindParam( 6, pool.getLong("maxpintime"));
  stmt.bindParam( 7, pool.getString("fss_policy"));
  stmt.bindParam( 8, pool.getString("gc_policy"));
  stmt.bindParam( 9, pool.getString("mig_policy"));
  stmt.bindParam(10, pool.getString("rs_policy"));
  stmt.bindParam(11, gids.str());
  stmt.bindParam(12, pool.getString("ret_policy"));
  stmt.bindParam(13, pool.getString("s_type"));
  stmt.bindParam(14, pool.type);
  stmt.bindParam(15, pool.serialize());
  stmt.bindParam(16, pool.name);
  
  if (stmt.execute() == 0)
    throw DmException(DM_NO_SUCH_POOL, "Pool %s not found", pool.name.c_str());
  
  // Tell the driver, since it may need to know
  driver->update(pool);
}



void MySqlPoolManager::deletePool(const Pool& pool) throw (DmException)
{
  // Notify the driver
  this->stack_->getPoolDriver(pool.type)->toBeDeleted(pool);
  
  // Remove from the database, ignoring if it does not exist, since
  // the handler may have done it (fs)
  Statement stmt(this->conn_, this->dpmDb_, STMT_DELETE_POOL);
  stmt.bindParam(0, pool.name);
  stmt.execute();
}



Location MySqlPoolManager::whereToRead(const std::string& path) throw (DmException)
{
  return this->whereToRead(this->stack_->getCatalog()->getReplicas(path));
}



Location MySqlPoolManager::whereToRead(ino_t inode) throw (DmException)
{
  return this->whereToRead(this->stack_->getINode()->getReplicas(inode));
}



Location MySqlPoolManager::whereToRead(const std::vector<Replica>& replicas) throw (DmException)
{
  unsigned i;
  std::vector<Location> available;
  
  if (replicas.size() == 0)
    throw DmException(DM_NO_REPLICAS, "No replicas");

  // See if available, and translate
  for (i = 0; i < replicas.size(); ++i) {
    if (replicas[i].hasField("pool")) {
      Pool pool = this->getPool(replicas[i].getString("pool"));
      
      try {
        PoolHandler* handler = this->stack_->getPoolDriver(pool.type)->
                                  createPoolHandler(pool.name);
        if (handler->replicaIsAvailable(replicas[i]))
          available.push_back(handler->whereToRead(replicas[i]));

        delete handler;
      }
      catch (DmException& e) {
        if (e.code() != DM_NO_SUCH_POOL) throw;
      }
    }
  }
  
  // Pick a random one from the available
  if (available.size() > 0) {
    i = rand() % available.size();
    return available[i];
  }
  else {
    throw DmException(DM_NO_REPLICAS,
                      "None of the replicas is available for reading");
  }
}



Location MySqlPoolManager::whereToWrite(const std::string& path) throw (DmException)
{
  std::vector<Pool> pools = this->getPools(PoolManager::kForWrite);
  if (pools.size() == 0)
    throw DmException(DM_NO_SPACE_LEFT, "There are no pools available for writing");
  
  // Pick a random one
  unsigned i = rand()  % pools.size();
  
  // Get the handler
  PoolHandler* handler = this->stack_->getPoolDriver(pools[i].type)->createPoolHandler(pools[i].name);
  
  // Create the entry
  this->stack_->getCatalog()->create(path, 0777);
  
  // Delegate to it
  Location loc = handler->whereToWrite(path);
  delete handler;
  
  // Done!
  return loc;
}
